"""
This file contains a list of constants that represent tags for JSON objects.
"""

# REBEL & BACKWARD_CHAINING LABELS
TOOL_ID = "ID"
DESCRIPTION ="DESC"
PARAMS = "PARAMS"

INTERACTION = "QA"
EXAMPLE = "CASE"
CONVERSATION = "MEM"

QUESTION = "Q"
THOUGHT = "THOUGHT"
PROMPT = "P"
RESPONSE = "A"

# FLAT LABELS
L_DESCRIPTION ="DESC"
L_PARAMS = "PARAMS"
L_BOT_INSTRUCTIONS = "SYSTEM"

L_INTERACTION = "INTERACTION"
L_INTERACTIONS = "INTERACTIONS"
L_EXAMPLE = "EXAMPLE"
L_EXAMPLES = "EXAMPLES"
L_CONVERSATION = "CONVERSATION-HISTORY"

L_THOUGHT = "THOUGHT"
L_TOOL_LIST = "TOOLS"
L_TOOL = "TOOL"
L_TOOL_PICKER = "PICK-TOOL"
L_QUESTION = "QUESTION"
L_ANSWER = "ANSWER"
L_ANSWER_DATA = "DATA"
L_ANSWER_SUMMARY = "SUMMARY"

L_SPLIT = "SPLIT"
L_SPLIT_INSTRUCTIONS = "INSTRUCTIONS"

SUBQ_SPLITTER = "|||"

# Human-AI interaction
L_INT_HUMAN = "HUMAN-AI"
# AI-AI interaction
L_INT_AI = "AI-AI"

QUALITY = 0.2



